"""Constants for the europa_1400_tools package."""

from enum import Enum, IntEnum
from pathlib import Path
from typing import Optional

# Game Directories and Files

DATA_DIR = "Data"
DATA_PATH = Path(DATA_DIR)
A_GEB_DAT = "A_Geb.dat"
AGEB_PATH = DATA_PATH / A_GEB_DAT
A_OBJ_DAT = "A_Obj.dat"
AOBJ_PATH = DATA_PATH / A_OBJ_DAT

GFX_DIR = "gfx"
GFX_PATH = Path(GFX_DIR)
BMP_DIR = "Bmp"
GROUNDPLANS_DIR = "GroundPlans"
MAPS_DIR = "Maps"
CHARACTER_NAMES_DAT = "character_names.dat"
GILDE_ADD_ON_GERMAN_GFX = "Gilde_add_on_german.gfx"
GILDE_ADD_ON_GERMAN_GFX_PATH = GFX_PATH / GILDE_ADD_ON_GERMAN_GFX
NO_COMPRESSION_INI = "no_compression.ini"

MOVIE_DIR = "Movie"
MSX_DIR = "msx"

RESOURCES_DIR = "Resources"
RESOURCES_PATH = Path(RESOURCES_DIR)
ANIMATIONS_BIN = "animations.bin"
ANIMATIONS_PATH = RESOURCES_PATH / ANIMATIONS_BIN
FORMS_BIN0 = "forms.bin0"
FORMS_BIN0_PATH = RESOURCES_PATH / FORMS_BIN0
GROUPS_BIN = "groups.bin"
GROUPS_PATH = RESOURCES_PATH / GROUPS_BIN
OBJECTS_BIN = "objects.bin"
OBJECTS_PATH = RESOURCES_PATH / OBJECTS_BIN
SCENES_BIN = "scenes.bin"
SCENES_PATH = RESOURCES_PATH / SCENES_BIN
SCRIPTS_BIN = "scripts.bin"
SCRIPTS_PATH = RESOURCES_PATH / SCRIPTS_BIN
SCRIPTS_BIN1 = "scripts.bin1"
SCRIPTS_PATH1 = RESOURCES_PATH / SCRIPTS_BIN1
TEXBIN_GERMAN_BIN = "textbin_german.bin"
TEXBIN_GERMAN_PATH = RESOURCES_PATH / TEXBIN_GERMAN_BIN
TEXTURES_BIN = "textures.bin"
TEXTURES_PATH = RESOURCES_PATH / TEXTURES_BIN

SERVER_DIR = "server"

SFX_DIR = "sfx"
SFX_PATH = Path(SFX_DIR)

AI_METHOD_ADDON_INI = "ai_method_addon.ini"
ALCHEMY_DATA_INI = "alchemy_data.ini"
DIE_GILDE_ADD_ON_TEXT_DEF = "die_gilde_add_on_text.def"
GAME_INI = "game.ini"
GILDEGOLD_EXE = "GildeGold.exe"
GILDEGOLD_TL_EXE = "GildeGold_TL.exe"
INCLUDE_SFX_INI = "include_sfx.ini"
MOVEAHEAD_DLL = "moveahead.dll"

# Directory and File Categories

BIN_FILES = [
    ANIMATIONS_BIN,
    FORMS_BIN0,
    GROUPS_BIN,
    OBJECTS_BIN,
    SCENES_BIN,
    SCRIPTS_BIN,
    SCRIPTS_BIN1,
    TEXBIN_GERMAN_BIN,
    TEXTURES_BIN,
]

CONVERTIBLE_PATHS = [
    AOBJ_PATH,
    AGEB_PATH,
    GILDE_ADD_ON_GERMAN_GFX_PATH,
    GROUPS_PATH,
    OBJECTS_PATH,
    SCENES_PATH,
    SFX_PATH,
]

# Output Directories and Files

EXTRACTED_DIR = "extracted"
DECODED_DIR = "decoded"
CONVERTED_DIR = "converted"
OUTPUT_AGEB_DIR = "ageb"
OUTPUT_AOBJ_DIR = "aobj"
OUTPUT_SFX_DIR = "sfx"
OUTPUT_GFX_DIR = "gfx"
OUTPUT_GROUPS_DIR = "groups"
OUTPUT_ANIMATIONS_DIR = "animations"
OUTPUT_SCENES_DIR = "scenes"
OUTPUT_OBJECTS_DIR = "objects"
OUTPUT_TEXTURES_DIR = "textures"
MAPPED_ANIMATONS_PICKLE = "mapped_animations.pickle"
MISSING_PATHS_TXT = "missing_paths.txt"

# File Extensions

BGF_EXTENSION = ".bgf"
OBJ_EXTENSION = ".obj"
MTL_EXTENSION = ".mtl"
BAF_EXTENSION = ".baf"
GLTF_EXTENSION = ".gltf"
GLB_EXTENSION = ".glb"
PNG_EXTENSION = ".png"
INI_EXTENSION = ".ini"
SBF_EXTENSION = ".sbf"
PICKLE_EXTENSION = ".pickle"
WAV_EXTENSION = ".wav"
MP3_EXTENSION = ".mp3"
GFX_EXTENSION = ".gfx"
LFS_EXTENSION = ".lfs"
OGR_EXTENSION = ".ogr"
JSON_EXTENSION = ".json"
OAM_EXTENSION = ".oam"
ED3_EXTENSION = ".ed3"
TXS_EXTENSION = ".txs"
DAT_EXTENSION = ".dat"
BIN_EXTENSION = ".bin"
TXT_EXTENSION = ".txt"

IGNORED_EXTENSIONS = [
    LFS_EXTENSION,
    TXS_EXTENSION,
    TXT_EXTENSION,
]

# Construct


class SoundbankType(IntEnum):
    """Soundbank types."""

    SINGLE = 1
    MULTI = 2


class SoundType(IntEnum):
    """Sbf types."""

    WAV = 1
    MP3 = 2


# Converter


class Format(Enum):
    """Formats."""

    def __str__(self) -> str:
        return self.value[0]

    @property
    def extension(self) -> str:
        return self.value[1]


class TyperTargetFormat(Enum):
    """Typer target formats."""

    WAV = "wav"
    MP3 = "mp3"
    WAVEFRONT = "wavefront"
    GLTF = "gltf"
    GLTF_STATIC = "gltf-static"
    JSON = "json"
    PNG = "png"


class TargetFormat(Format):
    """Target formats."""

    WAV = ("wav", WAV_EXTENSION)
    MP3 = ("mp3", MP3_EXTENSION)
    WAVEFRONT = ("wavefront", OBJ_EXTENSION)
    GLTF = ("gltf", GLTF_EXTENSION)
    GLTF_STATIC = ("gltf-static", GLTF_EXTENSION)
    JSON = ("json", JSON_EXTENSION)
    PNG = ("png", PNG_EXTENSION)

    @classmethod
    def convert_typer(cls):
        """Return an Enum type with values as the first elements of the tuples."""
        enum_dict = {format.name: format.value[0] for format in cls}
        return Enum("TyperTargetFormat", enum_dict)

    @staticmethod
    def from_typer(typer_target_format: str) -> Optional["TargetFormat"]:
        """Return the target format for the given typer target format."""
        for target_format in TargetFormat:
            if target_format.value[0] == typer_target_format:
                return target_format

        return None


class SourceFormat(Format):
    """Source formats."""

    BAF: tuple[str, str, list[TargetFormat], str, Path] = (
        "baf",
        BAF_EXTENSION,
        [],
        "animations",
        ANIMATIONS_PATH,
    )
    BGF: tuple[str, str, list[TargetFormat], str, Path] = (
        "bgf",
        BGF_EXTENSION,
        [
            TargetFormat.WAVEFRONT,
            TargetFormat.GLTF,
            TargetFormat.GLTF_STATIC,
        ],
        "objects",
        OBJECTS_PATH,
    )
    GFX: tuple[str, str, list[TargetFormat], str, Path] = (
        "gfx",
        GFX_EXTENSION,
        [
            TargetFormat.PNG,
        ],
        "graphics",
        GFX_PATH,
    )
    SBF: tuple[str, str, list[TargetFormat], str, Path] = (
        "sbf",
        SBF_EXTENSION,
        [
            TargetFormat.WAV,
            TargetFormat.MP3,
        ],
        "sounds",
        SFX_PATH,
    )
    OGR: tuple[str, str, list[TargetFormat], str, Path] = (
        "ogr",
        OGR_EXTENSION,
        [
            TargetFormat.JSON,
        ],
        "groups",
        GROUPS_PATH,
    )
    ED3: tuple[str, str, list[TargetFormat], str, Path] = (
        "ed3",
        ED3_EXTENSION,
        [
            TargetFormat.JSON,
        ],
        "scenes",
        SCENES_PATH,
    )
    AOBJ: tuple[str, str, list[TargetFormat], str, Path] = (
        "aobj",
        DAT_EXTENSION,
        [
            TargetFormat.JSON,
        ],
        "object_data",
        AOBJ_PATH,
    )
    AGEB: tuple[str, str, list[TargetFormat], str, Path] = (
        "ageb",
        DAT_EXTENSION,
        [
            TargetFormat.JSON,
        ],
        "building_data",
        AGEB_PATH,
    )

    @property
    def target_formats(self) -> list[TargetFormat]:
        return self.value[2]

    @property
    def command(self) -> str:
        return self.value[3]

    @property
    def source_path(self) -> Path | None:
        return self.value[4]

    @staticmethod
    def from_path(path: Path) -> Optional["SourceFormat"]:
        """Return the source format for the given extension."""

        for source_format in SourceFormat:
            if source_format.source_path is not None and str(path).endswith(
                str(source_format.source_path)
            ):
                return source_format

        for source_format in SourceFormat:
            if source_format.value[1].lower() == path.suffix.lower():
                return source_format

        return None


class OgrElementType(Enum):
    """Ogr element types."""

    OBJECT = "object"
    LIGHT = "light"
    DUMMY = "dummy"


BGF_DIR = "bgf"
OBJ_DIR = "obj"
TEX_DIR = "tex"
GLTF_DIR = "gltf"
BAF_DIR = "baf"

OBJECTS_STRING_ENCODING = "latin-1"

MODELS_REDUCED_FOOTER_FILES = [
    "ob_DREIFACHGALGEN.bgf",
    "ob_DREIFACKREUZ.bgf",
    "ob_EXEKUTIONSKANONESTOPFER.bgf",
]

BGF_EXCLUDE: list[Path] = []
BAF_EXCLUDE: list[Path] = []

WAVEFRONT_ENCODING = "utf-8"

BAF_INI_FILE_SECTION = "4HEAD Studios Animation-Settings"
BAF_INI_FILE_NUM_KEYS = "NumKeys"
BAF_INI_FILE_KEYS = "Keys"
BAF_INI_FILE_LOOP_IN = "LoopIn"
BAF_INI_FILE_LOOP_OUT = "LoopOut"


DEFAULT_OUTPUT_PATH = Path("output")


BAF_TO_BGFS = {
    "LowPolyCharacter/Soldat/karrenziehen_SOLDAT_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Soldat/stehen_SOLDAT_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Soldat/gehen_SOLDAT_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Junge/karrenziehen_JUNGE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Junge/gehen_JUNGE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Junge/stehen_JUNGE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Frau/karrenziehen_FRAU_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Frau/gehen_FRAU_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Frau/stehen_FRAU_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Mann/gehen_MANN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Mann/stehen_MANN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Mann/karrenziehen_MANN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Maedchen/gehen_MAEDCHEN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Maedchen/karrenziehen_MAEDCHEN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Maedchen/stehen_MAEDCHEN_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Mann2/karrenziehen_MANN2_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Mann2/gehen_MANN2_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Mann2/stehen_MANN2_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Kutte/gehen_KUTTE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Kutte/karrenziehen_KUTTE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Kutte/stehen_KUTTE_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Maedchen2_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin2_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiterin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Baeuerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin3_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen3_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Maedchen_Low_MAEDCHEN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Buergerin_Low_FRAU.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb2_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Domherr_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb4_Low_KUTTE.bgf",
        "_DYNAMIC/Low_Poly_Character/Abt_Low_KUTTE.bgf",
    ],
    "LowPolyCharacter/Dicker/karrenziehen_DICKER_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Dicker/gehen_DICKER_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "LowPolyCharacter/Dicker/stehen_DICKER_LOW.baf": [
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge2_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Patrizier_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Fechter_Low_SOLDAT.bgf",
        "_DYNAMIC/Low_Poly_Character/Zigeuner_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bettler2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Arbeiter_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Junge3_Low_JUNGE.bgf",
        "_DYNAMIC/Low_Poly_Character/Dicker_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Bauer_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb_Low_MANN2.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger3_Low_DICKER.bgf",
        "_DYNAMIC/Low_Poly_Character/Bergmann2_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Buerger_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Dieb3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Wirt3_Low_MANN.bgf",
        "_DYNAMIC/Low_Poly_Character/Armbrustschuetze_Low_SOLDAT.bgf",
    ],
    "Sonstiges/WIEGE1.BAF": ["Accessoires/Moebel/ub_WIEGE.bgf"],
    "Sonstiges/windmuehle.baf": [],
    "Sonstiges/sp_WIMPEL.baf": ["_DYNAMIC/X_STUFF/sp_WIMPEL.bgf"],
    "Sonstiges/diebesgilde_VORHANG.baf": [
        "Diebesgilde/Aufenthalt/ob_VORHANG_FUER_CHEF.bgf"
    ],
    "Sonstiges/diebesgilde_ILLUSIONSWAND_zu.baf": [
        "Diebesgilde/Gangsystem/ob_ILLUSIONSWAND.bgf",
        "Diebesgilde/Gefaengnis/ub_BOGEN_T_GEF_DIEBESGILDE.bgf",
    ],
    "Sonstiges/ub_Blutlache.baf": [
        "_DYNAMIC/Kampf/ub_Blutlache.bgf",
        "Schmiede/ub_STUMPF_01_SCHMIEDE.bgf",
        "Schmiede/ub_STUMPF_02_SCHMIEDE.bgf",
        "_Addon/Schneiderei/ub_SCHUERZE_B.bgf",
        "_Addon/Schneiderei/ub_STOFFBALLEN.bgf",
        "_Addon/Schneiderei/ub_SCHUERZE_A.bgf",
    ],
    "Sonstiges/anim_skelett.baf": [],
    "Sonstiges/ub_FLAGGENMAST.baf": [
        "Stadtwache/WACHZIMMER/ub_WAND_A_B_STWACHE_WACHZIMMER.bgf",
        "Accessoires/Aussen/ub_FLAGGENMAST.bgf",
        "Kerker/Folterkammer/ub_TISCH_A_GEF_FOLTERKAMMER.bgf",
        "Kerker/Folterkammer/ub_TISCH_B_GEF_FOLTERKAMMER.bgf",
    ],
    "Sonstiges/anim_FLUCHTFLAGGE.baf": ["_DYNAMIC/Kampf/ub_FLUCHTFLAGGE.bgf"],
    "Sonstiges/Wimpel_Stadtturm_K.baf": [
        "_DYNAMIC/X_STUFF/sp_STADTTURM_K.bgf",
        "_DYNAMIC/X_STUFF/sp_STADTTURM.bgf",
    ],
    "Sonstiges/STADTPUNKT_ANM_K.baf": ["_DYNAMIC/X_STUFF/sp_STADTPUNKT_K.bgf"],
    "Sonstiges/diebesgilde_ILLUSIONSWAND_auf.baf": [
        "Diebesgilde/Gangsystem/ob_ILLUSIONSWAND.bgf",
        "Diebesgilde/Gefaengnis/ub_BOGEN_T_GEF_DIEBESGILDE.bgf",
    ],
    "Sonstiges/WAAGE_PARFUMERIE_eins.baf": ["Parfuemerie/ob_WAAGE_PARFUMERIE.bgf"],
    "Sonstiges/Stadtturm_Anm_K.baf": [
        "_DYNAMIC/X_STUFF/sp_STADTTURM_K.bgf",
        "_DYNAMIC/X_STUFF/sp_STADTTURM.bgf",
    ],
    "Sonstiges/Wimpel_Stadtturm_K-V.baf": [
        "_DYNAMIC/X_STUFF/sp_STADTTURM_K.bgf",
        "_DYNAMIC/X_STUFF/sp_STADTTURM.bgf",
    ],
    "Sonstiges/Wimpel_Stadtturm.baf": [
        "_DYNAMIC/X_STUFF/sp_STADTTURM_K.bgf",
        "_DYNAMIC/X_STUFF/sp_STADTTURM.bgf",
    ],
    "Sonstiges/ob_ARMBRUST_hnd.baf": [
        "_DYNAMIC/Waffen/ob_ARMBRUST_A.bgf",
        "Accessoires/Kisten_usw/ob_WASSERBOTTICH.bgf",
        "_Addon/Friedhof/Kapelle/ob_OPFERSTEIN.bgf",
    ],
    "Sonstiges/ob_SCHRAUBPRESSE_PARFUMERIE.baf": [
        "Parfuemerie/ob_SCHRAUBPRESSE_PARFUMERIE.bgf"
    ],
    "Vogel/vogel_TAUBE.baf": ["_DYNAMIC/Tiere/vogel_TAUBE.bgf"],
    "Buch/Buch_zurueckblaettern.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "Buch/Buch_schliessen.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "Buch/Buch_schliessen_s.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "Buch/Buch_vorblaettern.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "Buch/Buch_zurueckblaettern_s.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "Buch/Buch_vorblaettern_s.baf": ["_DYNAMIC/Buch/Buch.bgf"],
    "_ADDON/AO_Kanone_Wimpel_Aa.baf": [
        "_Addon/_DYNAMIC/AO_Kanone_Wimpel_Aa.bgf",
        "_Addon/Domizil/Geheimkabinett/TREPPE_GEHEIM_Aa.bgf",
    ],
    "_ADDON/AO_Mech_Geist_Aa.baf": ["_Addon/Friedhof/Krypta/ob_MECHANISCHER_GEIST.bgf"],
    "_ADDON/AO_WIMPEL_ZIRKUSZELT.baf": [
        "_Addon/Domizil/Geheimkabinett/QUERBALKEN_GEHEIM_Ao.bgf",
        "_Addon/Zigeunerlager/ub_WIMPEL_ZIRKUSZELT.bgf",
    ],
    "Character/MANN/Stand_still_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Stand_still_MANN_BAK.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/ruecksalto_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/tief_verbeugen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/posieren_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/lachen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/handstand2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/stehend_beschwoeren_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/ruecksalto2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/anfeuern_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/lang_winken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ZIRKUS/handstand_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/naehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/muenz_praegung_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/metall_erhitzen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/hobeln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/schmieden_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/hacken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/umruehren_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/meisseln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/schweissab_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/neufegen2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/zapfen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/seil_ziehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/mist_schaufeln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/schoepfen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/baumfaellen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/hebel_ziehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/am_tisch_schreiben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/umruehren_auftisch_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/schreiben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/ARBEIT/saegen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/springen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/zu_boden_sinken_MANN.baf": [],
    "Character/MANN/Bewegung/karren_ziehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/arbeiter_KARREN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/holzfaeller_KARREN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker_KARREN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/gehen6_MANN.baf": [],
    "Character/MANN/Bewegung/dreh_90_rechts_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/manipulieren_hockend_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/step_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/leitersteigen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/gebueckt_gehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/gehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/dreh_45_rechts_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/dreh_90_links_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Bewegung/dreh_45_links_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_still_MANN.baf": [],
    "Character/MANN/SITZEND/sitz_kopfdreh_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_reden_aufgeregt_MANN.baf": [],
    "Character/MANN/SITZEND/sitz_schreiben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_noise1_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_noise3_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_nein_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_noise2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_kopfsenk_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/aufstehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitzgespreach3_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_manipulieren_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_umschau_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_trinken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/hinsetzen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_newnoise_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_reden_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_noise4_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitzgespreach2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SITZEND/sitz_ja_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/Kritzelkratzel/offizier_SOLDAT_stich_ausweichen.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/MANN/Kritzelkratzel/offizier_SOLDAT_schmieden.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/MANN/Kritzelkratzel/handwerker2_MANN_am_tisch_schreiben.baf": [],
    "Character/MANN/Kritzelkratzel/handwerker2_MANN_hobeln.baf": [],
    "Character/MANN/Kritzelkratzel/handwerker2_MANN_einhand_vom_tisch.baf": [],
    "Character/MANN/Kritzelkratzel/handwerker2_MANN_saegen.baf": [],
    "Character/MANN/Kritzelkratzel/handwerker2_MANN_vom_boden_nehmen.baf": [],
    "Character/MANN/Kritzelkratzel/tischnehmen1_MANN.baf": [],
    "Character/MANN/Kritzelkratzel/offizier_SOLDAT_werfen.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/MANN/Kritzelkratzel/ratsherr2_KUTTE_predigen.baf": [],
    "Character/MANN/KAMPF/todumfallen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/zu_boden_sinken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/hieb_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schlagen3_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/boxen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/hieb_abwehr_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/ducken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/wegstossen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schulter_anrempeln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/ausweichen_seite_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/aufstehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/priester_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schlagen2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/streich_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/budspencerhieb_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/stich_abwehr_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schlagen1_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schiessen_pistole_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/bauch_greifen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/sterben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/stich_ausweichen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/arm_greifen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/stich_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/aufrichten_MANN.baf": [],
    "Character/MANN/KAMPF/ausweichen_hinten_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/taumeln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/schiessen_armbrust_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/stechen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/KAMPF/werfen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/verbeugen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/armsehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kopf_schuetteln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kinn_kratzen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/durch_haar_fahr_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/herranwinken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/steh_noise7_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/haende_reiben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/nach_unten_sehen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/gespraech6_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/manipulieren_hockend_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/gespraech5_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/manipulieren_stehend_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kopf_schraeg_rechts_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/abfaellig_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/mit_faust_droh_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/zeigen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/tresen_wischen_MANN.baf": [],
    "Character/MANN/GESTIK/um_hand_anhalten_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/gespraech1_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kopf_nicken_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/steh_noise8_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/bauch_reiben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kopf_schraeg_links_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/bein_kratzen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/klatschen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/anpreisen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/segnen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/jubeln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/zum_mund_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/kopf_kratzen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/steh_noise9_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/beschwoeren_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/GESTIK/steh_noise61_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/abstimmung/sitzung3_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/abstimmung/sitzung4_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/abstimmung/sitzung2_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/abstimmung/sitzung1_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/abstimmung/sitzung5_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/STEHEN/stehen_1_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_7_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_4_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_3_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_2_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_newnoise_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/STEHEN/stehen_6_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_9_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_5_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/STEHEN/stehen_6a_MANN.baf": [],
    "Character/MANN/STEHEN/stehen_8_MANN.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN/SPEZIAL/meckern_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/singen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/runterbeug_such_MANN.baf": [],
    "Character/MANN/SPEZIAL/laterne_anzuenden_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/zweihand_vom_boden_aufnehmen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/fallenlassen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/auskippen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/wuerfel_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/niederkniehen_ehre_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/angeln_MANN.baf": [],
    "Character/MANN/SPEZIAL/gegenstand_schuetteln_zweihaendig_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/einhand_vom_tisch_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/runterbeug_sucht_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/in_guertel_steck_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/niederkniehen_tod_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/einhaendig_vom_boden_nehmen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/zweihand_vom_tisch_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/auf_ruecken_schwing_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/Scheiterhaufen.baf": [],
    "Character/MANN/SPEZIAL/in_andere_hand_geben_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/auf_tisch_legen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/in_tasche_steck_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/Scheiterhaufen_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN/SPEZIAL/gegenstand_schuetteln_MANN.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/Huhn/Bewegung/fressen_HUHN.baf": ["_DYNAMIC/Tiere/huhn_HUHN.bgf"],
    "Character/Huhn/Bewegung/gehen_HUHN.baf": ["_DYNAMIC/Tiere/huhn_HUHN.bgf"],
    "Character/Huhn/Bewegung/stehen_HUHN.baf": ["_DYNAMIC/Tiere/huhn_HUHN.bgf"],
    "Character/Huhn/Bewegung/flattern_HUHN.baf": ["_DYNAMIC/Tiere/huhn_HUHN.bgf"],
    "Character/BAER/warten_noise_BAER.baf": [
        "_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"
    ],
    "Character/BAER/Tanzen_BAER.baf": ["_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"],
    "Character/BAER/warten_BAER.baf": ["_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"],
    "Character/BAER/Hinsetzen_BAER.baf": [
        "_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"
    ],
    "Character/BAER/Aufrichten_BAER.baf": [
        "_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"
    ],
    "Character/BAER/Tanzen_im_Kreis_BAER.baf": [
        "_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"
    ],
    "Character/BAER/Laufen_BAER.baf": ["_Addon/_DYNAMIC/Character/Tanzbaer_BAER.bgf"],
    "Character/Ratte/Bewegung/stehen_RATTE.baf": ["_DYNAMIC/Tiere/ratte_RATTE.bgf"],
    "Character/Ratte/Bewegung/gehen_RATTE.baf": ["_DYNAMIC/Tiere/ratte_RATTE.bgf"],
    "Character/Ratte/Bewegung/fressen_RATTE.baf": ["_DYNAMIC/Tiere/ratte_RATTE.bgf"],
    "Character/DICKER/Stand_still_DICKER.baf": [],
    "Character/DICKER/Master_Dicker.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/hebel_ziehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/baumfaellen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/neufegen2_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/seil_ziehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/hobeln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/schmieden_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/metall_erhitzen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/naehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/saegen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/zapfen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/umruehren_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/schweissab_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/muenz_praegung_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/mist_schaufeln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/hacken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/am_tisch_schreiben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/meisseln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/schoepfen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/schreiben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/ARBEIT/umruehren_auftisch_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/springen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/step_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/dreh_90_rechts_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/dreh_45_rechts_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/leitersteigen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/zu_Boden_sinken_DICKER.baf": [],
    "Character/DICKER/Bewegung/gebueckt_gehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/karren_ziehen__DICKER.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/DICKER/Bewegung/dreh_45_links_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/dreh_90_links_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/gehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Bewegung/Karren_ziehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_reden_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_schreiben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_ja_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitzgespraech3_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_reden_aufgeregt_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_newnoise_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/hinsetzen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_noise1_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitzgespraech2_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_kopfsenk_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_still_DICKER.baf": [],
    "Character/DICKER/SITZEND/sitz_manipulieren_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_nein_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_kopfdreh_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/aufstehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_noise2_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_umschau_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_trinken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_noise4_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SITZEND/sitz_noise3_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/bauch_greifen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/stich_ausweichen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/todumfallen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/ducken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/wegstossen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schulter_anrempeln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/werfen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schiessen_pistole_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/priester_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/stechen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/boxen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schlagen2_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/budspencerhieb_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/ausweichen_seite_DICKER.baf": [],
    "Character/DICKER/Kampf/stich_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/sterben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/streich_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schlagen3_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/aufstehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schiessen_armbrust_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/taumeln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/hieb_abwehr_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/hieb_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/arm_greifen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/schlagen1_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/stich_abwehr_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/zu_boden_sinken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/Kampf/ausweichen_hinten_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/herranwinken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/armsehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kopf_schraeg_links_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/jubeln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/zeigen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/steh_noise61_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/durch_haar_fahr_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/steh_noise8_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/steh_noise9_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/gespraech5_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kinn_kratzen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/abfaellig_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/mit_faust_droh_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/manipulieren_stehend_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kopf_nicken_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/klatschen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/steh_noise7_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/nach_unten_sehen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/zum_mund_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/beschwoeren_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/anpreisen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kopf_kratzen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/manipulieren_hockend_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/gespraech6_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/haende_reiben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/um_hand_anhalten_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/bauch_reiben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/verbeugen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/segnen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/bein_kratzen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kopf_schuetteln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/kopf_schraeg_rechts_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/GESTIK/gespraech1_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/abstimmung/sitzung3_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/abstimmung/sitzung1_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/abstimmung/sitzung4_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/abstimmung/sitzung2_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/abstimmung/sitzung5_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/STEHEN/stehen_newnoise_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/in_tasche_steck_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/zweihand_vom_boden_aufnehmen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/betrunken_DICKER.baf": [],
    "Character/DICKER/SPEZIAL/in_guertel_steck_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/singen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/laterne_anzuenden_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/niederkniehen_tod_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/einhand_vom_tisch_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/scheiterhaufen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/einhaendig_vom_boden_nehmen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/meckern_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/gegenstand_schuetteln_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/runterbeug_such_DICKER.baf": [],
    "Character/DICKER/SPEZIAL/auskippen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/niederkniehen_ehre_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/zweihand_vom_tisch_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/fallenlassen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/auf_tisch_legen_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/wuerfel_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/auf_ruecken_schwing_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/gegenstand_schuetteln_zweihaendig_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/runterbeug_sucht_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/DICKER/SPEZIAL/taumelnd_DICKER.baf": [],
    "Character/DICKER/SPEZIAL/in_andere_hand_geben_DICKER.baf": [
        "_DYNAMIC/Character/Wirt2_DICKER.bgf",
        "_DYNAMIC/Character/wirt_DICKER.bgf",
        "_DYNAMIC/Character/Buerger3_DICKER.bgf",
        "_DYNAMIC/Character/schneider3_DICKER.bgf",
        "_DYNAMIC/Character/Dicker_DICKER.bgf",
    ],
    "Character/MANN3/Spezial/scheiterhaufen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/dreh_90_links_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/dreh_45_links_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/dreh_45_rechts_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/dreh_90_rechts_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/stehen_newnoise_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Bewegung/gehen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Abstimmung/sitzung4_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Abstimmung/sitzung2_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Abstimmung/sitzung3_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Abstimmung/sitzung1_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Abstimmung/sitzung5_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/steh_noise9_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/kopf_schraeg_rechts_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/gespraech1_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/abfaellig_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/durch_haar_fahr_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/bein_kratzen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/mit_faust_droh_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/steh_noise7_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/kopf_nicken_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/kopf_schuetteln_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/bauch_reiben_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/steh_noise8_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/nach_unten_sehen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/haende_reiben_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/kinn_kratzen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Gestik/gespraech5_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Kampf/stich_ausweichen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Kampf/taumeln_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Kampf/schiessen_pistole_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_ja_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_reden_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_noise1_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/hinsetzen_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_nein_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_newnoise_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitz_kopfsenk_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitzgespraech2_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/sitzend/sitzgespraech3_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MANN3/Stehen/stehen_newnoise_MANN3.baf": [
        "_Addon/_DYNAMIC/Character/ZIGEUNERMEISTER_MANN3.bgf"
    ],
    "Character/MEISTER/Spezial/niederkniehen_tod_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Spezial/scheiterhaufen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Bewegung/dreh_45_links_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Bewegung/dreh_90_rechts_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Bewegung/dreh_90_links_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Bewegung/gehen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Bewegung/dreh_45_rechts_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Abstimmung/sitzung5_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Abstimmung/sitzung1_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Abstimmung/sitzung3_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Abstimmung/sitzung2_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Abstimmung/sitzung4_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/mit_faust_droh_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/kinn_kratzen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/durch_haar_fahr_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/steh_noise9_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/nach_unten_sehen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/kopf_schuetteln_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/gespraech1_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/abfaellig_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/haende_reiben_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/bauch_reiben_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/gespraech5_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/steh_noise8_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/beschwoeren_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/steh_noise7_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/stehen_noise61_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/gespraech6_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Gestik/kopf_nicken_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Kampf/stich_ausweichen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Kampf/taumeln_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/Kampf/schiessen_pistole_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/hinsetzen_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_reden_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitzgespraech2_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_noise1_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_kopfsenk_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_nein_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitzgespraech3_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_ja_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/sitzend/sitz_newnoise_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/MEISTER/STEHEN/stehen_newnoise_MEISTER.baf": [
        "_Addon/_DYNAMIC/Character/Nekromant_MEISTER.bgf"
    ],
    "Character/medium/Bewegung/dreh_45_rechts_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/medium/Bewegung/gehen_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/medium/Bewegung/dreh_45_links_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/medium/Bewegung/dreh_90_links_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/medium/Bewegung/karren_ziehen_MEDIUM.baf": [],
    "Character/medium/Bewegung/dreh_90_rechts_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/medium/STEHEN/stehen_newnoise_MEDIUM.baf": [
        "_DYNAMIC/Character/Spion2_MEDIUM.bgf",
        "_DYNAMIC/Character/Spion_MEDIUM.bgf",
    ],
    "Character/karren/Bewegung/karren_ziehen_KARREN.baf": [
        "_DYNAMIC/Character/arbeiter_KARREN.bgf",
        "_DYNAMIC/Character/holzfaeller_KARREN.bgf",
        "_DYNAMIC/Character/handwerker_KARREN.bgf",
    ],
    "Character/MANN2/Stand_still_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/mist_schaufeln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/baumfaellen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/hobeln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/naehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/hacken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/schoepfen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/meisseln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/seil_ziehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/muenz_praegung_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/saegen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/schweissab_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/umruehren_auftisch_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/neufegen2_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/schreiben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/zapfen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/hebel_ziehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/umruehren_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/metall_erhitzen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/am_tisch_schreiben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/ARBEIT/schmieden_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/dreh_45_rechts_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/gehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/zu_boden_sinken_MANN2.baf": [],
    "Character/MANN2/Bewegung/leitersteigen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/karren_ziehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/arbeiter_KARREN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/holzfaeller_KARREN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker_KARREN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/gebueckt_gehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/dreh_90_rechts_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/dreh_90_links_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/springen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/dreh_45_links_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/Bewegung/step_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_manipulieren_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_noise3_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_still_MANN2.baf": [],
    "Character/MANN2/SITZEND/sitz_schreiben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/aufstehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_kopfdreh_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_newnoise_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_noise1_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_ja_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/hinsetzen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_reden_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_umschau_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_trinken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_nein_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_noise4_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_reden_aufgeregt_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_noise2_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SITZEND/sitz_kopfsenk_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/NEU/linksdrehen90_MANN.baf": [],
    "Character/MANN2/NEU/schritt_MANN.baf": [],
    "Character/MANN2/NEU/rechtsdrehen45_MANN.baf": [],
    "Character/MANN2/NEU/rechtsdrehen90_MANN.baf": [],
    "Character/MANN2/NEU/linksdrehen45_MANN.baf": [],
    "Character/MANN2/KAMPF/budspencerhieb_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/boxen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/arm_greifen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/aufrichten_MANN2.baf": [],
    "Character/MANN2/KAMPF/zu_boden_sinken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schiessen_pistole_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/stich_abwehr_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schlagen2_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/todumfallen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/ausweichen_hinten_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/aufstehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/taumeln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/ducken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/stich_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/streich_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/wegstossen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/hieb_abwehr_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/ausweichen_seite_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schulter_anrempeln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/sterben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/stechen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/priester_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schlagen1_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/werfen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/hieb_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/bauch_greifen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schlagen3_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/schiessen_armbrust_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/KAMPF/stich_ausweichen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kopf_schraeg_rechts_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/um_hand_anhalten_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/verbeugen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/steh_noise9_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/steh_noise8_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kopf_kratzen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/gespraech1_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/zeigen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kinn_kratzen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kopf_schuetteln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/abfaellig_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/manipulieren_hockend_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/bein_kratzen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/anpreisen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/gespraech5_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kopf_nicken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/zum_mund_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/gespraech6_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/nach_unten_sehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/zum_mund_MAN2.baf": [],
    "Character/MANN2/GESTIK/steh_noise7_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/kopf_schraeg_links_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/beschwoeren_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/jubeln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/steh_noise61_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/armsehen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/herranwinken_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/klatschen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/durch_haar_fahr_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/haende_reiben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/mit_faust_droh_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/manipulieren_stehend_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/bauch_reiben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/GESTIK/segnen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/abstimmung/sitzung3_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/abstimmung/sitzung4_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/abstimmung/sitzung1_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/abstimmung/sitzung5_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/abstimmung/sitzung2_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/STEHEN/stehen_6_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_5_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_6a_MANN2.baf": [],
    "Character/MANN2/STEHEN/stehen_1_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_7_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_3_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_2_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_8_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_4_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_9_MANN2.baf": [
        "Gebaeude/gb_Zunfthaus/gb_ZUNFTHAUS_0.bgf"
    ],
    "Character/MANN2/STEHEN/stehen_newnoise_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/fallenlassen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/gegenstand_schuetteln_zweihaendig_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/niederkniehen_tod_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/einhand_vom_tisch_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/in_tasche_steck_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/zweihand_vom_tisch_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/runterbeug_such_MANN2.baf": [],
    "Character/MANN2/SPEZIAL/in_guertel_steck_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/auskippen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/laterne_anzuenden_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/runterbeug_sucht_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/zweihand_vom_boden_aufnehmen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/niederkniehen_ehre_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/meckern_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/gegenstand_schuetteln_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/einhaendig_vom_boden_nehmen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/wuerfel_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/singen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/auf_tisch_legen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/in_andere_hand_geben_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/scheiterhaufen_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/MANN2/SPEZIAL/auf_ruecken_schwing_MANN2.baf": [
        "_DYNAMIC/Character/schneider1_MANN.bgf",
        "_DYNAMIC/Character/schneider2_MANN.bgf",
        "_DYNAMIC/Character/handwerker5_MANN.bgf",
        "_DYNAMIC/Character/zigeuner_MANN.bgf",
        "_DYNAMIC/Character/dieb_MANN2.bgf",
        "_DYNAMIC/Character/Arbeiter_MANN.bgf",
        "_DYNAMIC/Character/handwerker6_MANN.bgf",
        "_DYNAMIC/Character/Handwerker3_MANN.bgf",
        "_DYNAMIC/Character/wirt3_MANN.bgf",
        "_DYNAMIC/Character/Handwerker2_MANN.bgf",
        "_DYNAMIC/Character/HANDWERKER_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann2_MANN.bgf",
        "_DYNAMIC/Character/Bergmann_MANN.bgf",
        "_DYNAMIC/Character/buerger2_MANN.bgf",
        "_DYNAMIC/Character/buerger_MANN.bgf",
        "_DYNAMIC/Character/Bauer_MANN2.bgf",
        "_DYNAMIC/Character/Patrizier_MANN2.bgf",
        "_DYNAMIC/Character/Bergmann3_MANN.bgf",
        "_DYNAMIC/Character/stallbursche_MANN.bgf",
        "_DYNAMIC/Character/handwerker4_MANN.bgf",
        "_DYNAMIC/Character/Handwerker_MANN.bgf",
    ],
    "Character/Kritzelkratz/ratsherr2_KUTTE_gehen.baf": [],
    "Character/Kritzelkratz/offizier_SOLDAT_schiessen_pistolealt.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/Katze/Bewegung/stehen_KATZE.baf": ["_DYNAMIC/Tiere/katze_KATZE.bgf"],
    "Character/Katze/Bewegung/gehen_KATZE.baf": ["_DYNAMIC/Tiere/katze_KATZE.bgf"],
    "Character/KREATUR/Spezial/stehlen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Spezial/auferstehung_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Spezial/Elektroschock_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/dreh_45_links_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/gehen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/dreh_45_rechts_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/dreh_90_links_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/dreh_90_rechts_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Bewegung/aufstehen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/sterben_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/manipulieren_stehend_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/blume_im_mund_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/verbeugen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/ruelpser_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/schimpfen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/beschwoeren_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/umkippen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/manipulieren_hockend_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/schlaege_liegen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Gestik/schlaege_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Arbeit/neufegen2_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/Arbeit/buddeln_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/STEHEN/stehen_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/KREATUR/STEHEN/stehen_newnoise_KREATUR.baf": [
        "_Addon/_DYNAMIC/Character/Diener_KREATUR.bgf"
    ],
    "Character/Hund/Bewegung/stehen_HUND.baf": ["_DYNAMIC/Tiere/hund_HUND.bgf"],
    "Character/Hund/Bewegung/gehen_HUND.baf": ["_DYNAMIC/Tiere/hund_HUND.bgf"],
    "Character/Hund/Bewegung/bellen_HUND.baf": ["_DYNAMIC/Tiere/hund_HUND.bgf"],
    "Character/SOLDAT/Stand_still_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/zapfen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/schweissab_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/schreiben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/umruehren_auftisch_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/meisseln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/seil_ziehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/hacken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/saegen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/mist_schaufeln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/am_tisch_schreiben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/naehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/baumfaellen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/neufegen2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/hebel_ziehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/muenz_praegung_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/hobeln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/schoepfen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/umruehren_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/schmieden_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/ARBEIT/metall_erhitzen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/gebueckt_gehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/step_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/dreh_90_rechts_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/dreh_90_links_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/zu_boden_sinken_SOLDAT.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/springen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/dreh_45_links_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/leitersteigen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/dreh_45_rechts_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/gehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Bewegung/karren_ziehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise5_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_reden_aufgeregt_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech3_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_reden_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_kopfdreh_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/aufstehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise3_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise6_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/hinsetzen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_trinken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_nein_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise7_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech4_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_manipulieren_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech7_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech6_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech1_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise1_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_umschau_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzend_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise4_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise8_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_ja_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_greifen_SOLDAT.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitzgespraech5_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_schreiben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_still_SOLDAT.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_newnoise_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_kopfsenk_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SITZEND/sitz_noise2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/streich_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/boxen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/stechen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schiessen_armbrust_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/aufstehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/sterben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/zu_boden_sinken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/hieb_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/stich_ausweichen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schlagen1_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/todumfallen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/aufrichten_SOLDAT.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/taumeln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/ducken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/wegstossen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schlagen3_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/priester_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/ausweichen_seite_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/stich_abwehr_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/bauch_greifen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/ausweichen_hinten_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/arm_greifen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/stich_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/budspencerhieb_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schiessen_pistole_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/hieb_abwehr_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schulter_anrempeln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/schlagen2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/werfen2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/Kampf/werfen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kopf_schuetteln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/gespraech6_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/anpreisen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/klatschen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/abfaellig_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kopf_kratzen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/segnen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/gespraech5_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/zum_mund_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/herranwinken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kopf_schraeg_links_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/steh_noise61_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kopf_nicken_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/haende_reiben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/mit_faust_droh_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/um_hand_anhalten_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/manipulieren_hockend_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/bauch_reiben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/manipulieren_stehend_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/steh_noise8_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/durch_haar_fahr_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kopf_schraeg_rechts_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/kinn_kratzen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/gespraech1_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/nach_unten_sehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/beschwoeren_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/steh_noise9_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/armsehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/zeigen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/bein_kratzen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/steh_noise7_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/jubeln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/GESTIK/verbeugen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/abstimmung/sitzung1_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/abstimmung/sitzung3_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/abstimmung/sitzung5_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/abstimmung/sitzung2_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/abstimmung/sitzung4_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/STEHEN/stehen_newnoise_KUTTE.baf": [],
    "Character/SOLDAT/STEHEN/stehen_newnoise_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/in_andere_hand_geben_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/niederkniehen_tod_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/einhaendig_vom_boden_nehmen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/auskippen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/wuerfel_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/runterbeug_sucht_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/meckern_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/fallenlassen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/zweihand_vom_boden_aufnehmen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/niederkniehen_ehre_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/gegenstand_schuetteln_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/in_tasche_steck_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/runterbeug_such_SOLDAT.baf": [
        "_DYNAMIC/Character/armbrustschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/hellebardentraeger_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/auf_tisch_legen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/gegenstand_schuetteln_zweihaendig_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/auf_ruecken_schwing_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/singen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/einhand_vom_tisch_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/scheiterhaufen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/in_guertel_steck_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/laterne_anzuenden_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/SOLDAT/SPEZIAL/zweihand_vom_tisch_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/Schaf/Bewegung/fressen_SCHAF.baf": ["_DYNAMIC/Tiere/schaf_SCHAF.bgf"],
    "Character/Schaf/Bewegung/stehen_SCHAF.baf": ["_DYNAMIC/Tiere/schaf_SCHAF.bgf"],
    "Character/Schaf/Bewegung/gehen_SCHAF.baf": ["_DYNAMIC/Tiere/schaf_SCHAF.bgf"],
    "Character/Pferd/Bewegung/galopp_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/traben_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/fressen_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/wiehern_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/stehen_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/gehen_PFERD.baf": ["_DYNAMIC/Tiere/pferd_PFERD.bgf"],
    "Character/Pferd/Bewegung/karren_ziehen_PFERD.baf": [
        "_DYNAMIC/Tiere/pferd_PFERD.bgf"
    ],
    "Character/KUTTE2/Gestik/kinn_kratzen_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/KUTTE2/Gestik/Manipulieren_stehend_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/KUTTE2/bewegung/gehen_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/KUTTE2/bewegung/dreh_90_rechts_KUTTE2.baf": [
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uml_04.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Mauer_u_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uvl_02.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_umr_03.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uvr_01.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Mauer_o_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BUERGERHAUS_a/bau_BH_pfeiler_o_11.bgf",
        "Accessoires/Felsen/Felsen_07_doppel.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/DOMIZIL_b/bau_Burg_Stufe2_10.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/DOMIZIL_a/bau_Schloss_Stufe1_05.bgf",
    ],
    "Character/KUTTE2/bewegung/dreh_45_rechts_KUTTE2.baf": [
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uml_04.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Mauer_u_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uvl_02.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_umr_03.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_MITTEL_a/bauDomA_Turm_uvr_01.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Mauer_o_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BUERGERHAUS_a/bau_BH_pfeiler_o_11.bgf",
        "Accessoires/Felsen/Felsen_07_doppel.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/DOMIZIL_b/bau_Burg_Stufe2_10.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/DOMIZIL_a/bau_Schloss_Stufe1_05.bgf",
    ],
    "Character/KUTTE2/bewegung/dreh_45_links_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/KUTTE2/bewegung/stehen_newnoise_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/KUTTE2/bewegung/dreh_90_links_KUTTE2.baf": [
        "_Addon/_DYNAMIC/Character/Wissenschaftler_KUTTE2.bgf"
    ],
    "Character/FRAU3/Spezial/auferstehung_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Spezial/Elektroschock_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Spezial/Stehlen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/dreh_90_rechts_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/aufstehen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/dreh_90_links_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/dreh_45_rechts_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/dreh_45_links_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Bewegung/gehen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Sterben_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/manipulieren_stehend_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Schimpfen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Ruelpser_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Blume_im_Mund_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/schlaege_liegen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/schlaege_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Verbeugen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Manipulieren_Hockend_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Umkippen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Gestik/Beschwoeren_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Stehen/stehen_newnoise_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Stehen/stehen_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Arbeit/neufegen2_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/FRAU3/Arbeit/Buddeln_FRAU3.baf": [
        "_Addon/_DYNAMIC/Character/Friedhofsgehilfin_FRAU3.bgf"
    ],
    "Character/KUTTE/Bewegung/dreh_90_rechts_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/priester_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/gebueckt_gehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/dreh_45_rechts_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/zu_boden_sinken_KUTTE.baf": [],
    "Character/KUTTE/Bewegung/dreh_45_links_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/step_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/karren_ziehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/dreh_90_links_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/leitersteigen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/springen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Bewegung/gehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_kopfdreh_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_kopfsenk_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech5_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzend_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_trinken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/aufstehen_SOLDAT.baf": [
        "_DYNAMIC/Character/raeubermeister_SOLDAT.bgf",
        "_DYNAMIC/Character/musketenschuetze_SOLDAT.bgf",
        "_DYNAMIC/Character/raeuberlehrling_SOLDAT.bgf",
        "_DYNAMIC/Character/feldarbeiter_SOLDAT.bgf",
        "_DYNAMIC/Character/holzfaeller_SOLDAT.bgf",
        "_DYNAMIC/Character/offizier_SOLDAT.bgf",
        "_DYNAMIC/Character/raeubergeselle_SOLDAT.bgf",
        "_DYNAMIC/Character/ratsherr_SOLDAT.bgf",
        "_DYNAMIC/Character/fechter_SOLDAT.bgf",
        "_DYNAMIC/Character/kanonier_SOLDAT.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise3_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_newnoise_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/hinsetzen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech6_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise6_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_ja_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise1_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech7_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech1_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech4_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise8_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech3_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_umschau_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_reden_aufgeregt_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitzgespraech2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/aufstehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_reden_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_nein_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise7_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_still_KUTTE.baf": [],
    "Character/KUTTE/SITZEND/sitz_noise5_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_schreiben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_noise4_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SITZEND/sitz_manipulieren_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/stich_abwehr_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/stich_ausweichen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/priester_KUTTE.baf": [],
    "Character/KUTTE/KAMPF/werfen2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/budspencerhieb_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/werfen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schulter_anrempeln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/zu_boden_sinken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/bauch_greifen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/aufrichten_KUTTE.baf": [],
    "Character/KUTTE/KAMPF/taumeln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/wegstossen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/sterben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/stich_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schiessen_armbrust_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/todumfallen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/boxen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/streich_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/ausweichen_seite_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/hieb_abwehr_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schiessen_pistole_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/aufstehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/stechen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/arm_greifen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schlagen3_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schlagen1_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/ausweichen_hinten_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/schlagen2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/hieb_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/KAMPF/ducken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_kratzen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/jubeln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/gespraech6_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/steh_noise7_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/mit_faust_droh_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/haende_reiben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_schuetteln.baf": [],
    "Character/KUTTE/GESTIK/nach_unten_sehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/manipulieren_stehend_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/gespraech1_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/steh_noise8_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/segnen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/klatschen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_schraeg_links_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/bein_kratzen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kinn_kratzen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_schuetteln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/gespraech5_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/bauch_reiben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/zeigen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/beschwoeren_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/mit_faust_droh_,KUTTE.baf": [],
    "Character/KUTTE/GESTIK/kopf_schraeg_rechts_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/abfaellig_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/steh_noise61_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/anpreisen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/um_hand_anhalten_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/durch_haar_fahr_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/steh_noise9_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/armsehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_schraeg_links_UTTE.baf": [],
    "Character/KUTTE/GESTIK/herranwinken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/kopf_nicken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/zum_mund_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/manipulieren_hockend_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/GESTIK/verbeugen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/abstimmung/sitzung4_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/abstimmung/sitzung5_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/abstimmung/sitzung2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/abstimmung/sitzung1_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/abstimmung/sitzung3_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/am_tisch_schreiben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/schreiben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/schmieden_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/umruehren_auftisch_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/predigen_KUTTE.baf": [],
    "Character/KUTTE/Arbeit/saegen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/schweissab_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/schoepfen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/meisseln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/neufegen2_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/mist_schaufeln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/muenz_praegung_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/metall_erhitzen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/hacken_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/seil_ziehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/naehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/zapfen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/umruehren_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/hebel_ziehen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/baumfaellen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/Arbeit/hobeln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/STEHEN/stehen_newnoise_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/zweihand_vom_boden_aufnehmen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/laterne_anzuenden_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/runterbeug_sucht_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/gegenstand_schuetteln_zweihaendig_,KUTTE.baf": [],
    "Character/KUTTE/SPEZIAL/auf_tisch_legen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/einhaendig_vom_boden_nehmen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/einhand_vom_tisch_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/singen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/auf_ruecken_schwing_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/fallenlassen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/auskippen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/niederkniehen_tod_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/meckern_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/scheiterhaufen_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/runterbeug_such_KUTTE.baf": [],
    "Character/KUTTE/SPEZIAL/in_guertel_steck_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/gegenstand_schuetteln_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/in_tasche_steck_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/in_andere_hand_geben_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/wuerfel_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/gegenstand_schuetteln_zweihaendig_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/zweihand_vom_tisch_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/KUTTE/SPEZIAL/niederkniehen_ehre_KUTTE.baf": [
        "_DYNAMIC/Character/Ratsherr2_KUTTE.bgf",
        "_DYNAMIC/Character/Domherr_KUTTE.bgf",
        "_DYNAMIC/Character/Abt_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb4_KUTTE.bgf",
        "_DYNAMIC/Character/BETTLER2_KUTTE.bgf",
        "_DYNAMIC/Character/Priester_KUTTE.bgf",
        "_DYNAMIC/Character/Dieb2_KUTTE.bgf",
        "_DYNAMIC/Character/Bettler_KUTTE.bgf",
    ],
    "Character/Kuh/Bewegung/gehen_KUH.baf": ["_DYNAMIC/Tiere/kuh_KUH.bgf"],
    "Character/Kuh/Bewegung/stehen_KUH.baf": ["_DYNAMIC/Tiere/kuh_KUH.bgf"],
    "Character/Kuh/Bewegung/fressen_KUH.baf": ["_DYNAMIC/Tiere/kuh_KUH.bgf"],
    "Character/Rabe/umschauen_RABE.baf": ["_Addon/_DYNAMIC/Character/Rabe.bgf"],
    "Character/Rabe/warten_RABE.baf": ["_Addon/_DYNAMIC/Character/Rabe.bgf"],
    "Character/Rabe/kraehen_RABE.baf": ["_Addon/_DYNAMIC/Character/Rabe.bgf"],
    "Character/FRAU2/Sitzend/aufstehen_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/hinsetzen_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_schreiben_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_reden_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_newnoise_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_manipulieren_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_reden_aufgeregt_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Sitzend/sitz_trinken_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Bewegung/dreh_45_links_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Bewegung/dreh_90_links_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Bewegung/gehen_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Bewegung/dreh_45_rechts_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Bewegung/dreh_90_rechts_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Gestik/herranwinken_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Gestik/haende_reiben_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Gestik/kopf_nicken_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Gestik/beschwoeren_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Gestik/Manipulieren_stehend_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU2/Stehen/stehen_newnoise_FRAU2.baf": [
        "_Addon/_DYNAMIC/Character/Wahrsagerin_FRAU2.bgf"
    ],
    "Character/FRAU/Stand_still_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/posieren_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/handstand_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/lange_winken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/ruecksalto2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/anfeuern_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/lang_winken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/tief_verbeugen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/stehend_beschwoeren_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/ruecksalto_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/handstand2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ZIRKUS/lachen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/neufegen2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/hacken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/metall_erhitzen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/schweissab_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/schmieden_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/hobeln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/muenz_praegung_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/hebel_ziehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/zapfen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/umruehren_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/meisseln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/am_tisch_schreiben_FRAU.baf": [],
    "Character/FRAU/ARBEIT/schoepfen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/seil_ziehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/umruehren_auftisch_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/naehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/schreiben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/mist_schaufeln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/baumfaellen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/ARBEIT/saegen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/schritt_FRAU.baf": [],
    "Character/FRAU/Bewegung/rechtsdrehen90_FRAU.baf": [],
    "Character/FRAU/Bewegung/zu_boden_sinken_FRAU.baf": [
        "_DYNAMIC/Character/Holzfaellerin_FRAU.bgf"
    ],
    "Character/FRAU/Bewegung/dreh_45_rechts_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/rechtsdrehen45_FRAU.baf": [],
    "Character/FRAU/Bewegung/linksdrehen45_FRAU.baf": [],
    "Character/FRAU/Bewegung/dreh_90_links_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/gehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/karren_ziehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/linksdrehen90_FRAU.baf": [],
    "Character/FRAU/Bewegung/gebueckt_gehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/springen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/dreh_90_rechts_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/leitersteigen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/dreh_45_links_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Bewegung/step_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_schreiben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/hinsetzen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/aufstehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitzgespraech3_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_reden_aufgeregt_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_nein_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitzgespraech2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_noise1_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_umschau_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_ja_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_newnoise_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_trinken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_still_FRAU.baf": [
        "_DYNAMIC/Character/Holzfaellerin_FRAU.bgf"
    ],
    "Character/FRAU/SITZEND/sitz_reden_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_manipulieren_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_noise2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_noise3_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_noise4_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_kopfsenk_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SITZEND/sitz_kopfdreh_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/bauch_reiben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kopf_nicken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/steh_noise9_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/zeigen_zeigen_FRAU.baf": [
        "_DYNAMIC/Character/Holzfaellerin_FRAU.bgf"
    ],
    "Character/FRAU/Gestik/haende_reiben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/steh_noise61_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/durch_haar_fahr_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kopf_schuetteln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/abfaellig_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/herranwinken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/beschwoeren_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/anpreisen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/jubeln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/steh_noise8_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kinn_kratzen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kopf_schraeg_rechts_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kopf_kratzen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/nach_unten_sehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/steh_noise7_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/gespraech1_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/manipulieren_stehend_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/um_hand_anhalten_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/mit_faust_droh_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/segnen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/armsehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/kopf_schraeg_links_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/gespraech6_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/manipulieren_hockend_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/verbeugen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/bein_kratzen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/zeigen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/zum_mund_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/klatschen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/Gestik/gespraech5_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schlagen1_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/sterben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/stich_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/ducken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/budspencerhieb_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schlagen2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schiessen_armbrust_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/todumfallen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/zu_boden_sinken_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/aufstehen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/stechen_FRAU.baf": [],
    "Character/FRAU/KAMPF/stich_ausweichen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schulter_anrempeln_FRAU.baf": [],
    "Character/FRAU/KAMPF/ausweichen_hinten_FRAU.baf": [],
    "Character/FRAU/KAMPF/aufrichten_FRAU.baf": [
        "_DYNAMIC/Character/Holzfaellerin_FRAU.bgf"
    ],
    "Character/FRAU/KAMPF/ausweichen_seite_FRAU.baf": [],
    "Character/FRAU/KAMPF/stich_abwehr_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schlagen3_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/wegstossen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/priester_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/werfen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/streich_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/hieb_abwehr_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/hieb_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/boxen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/schiessen_pistole_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/arm_greifen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/bauch_greifen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/KAMPF/taumeln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/abstimmung/sitzung5_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/abstimmung/sitzung4_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/abstimmung/sitzung3_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/abstimmung/sitzung2_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/abstimmung/sitzung1_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/STEHEN/stehen_newnoise_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/in_andere_hand_geben_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/zweihand_vom_boden_aufnehmen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/niederkniehen_ehre_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/wuerfel_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/niederkniehen_tod_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/in_tasche_steck_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/auf_tisch_legen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/gegenstand_schuetteln_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/auf_ruecken_schwing_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/laterne_anzuenden_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/fallenlassen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/auskippen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/zweihand_vom_tisch_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/scheiterhaufen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/meckern_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/runterbeug_sucht_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/einhand_vom_tisch_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/in_guertel_steck_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/singen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/einhaendig_vom_boden_nehmen_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/FRAU/SPEZIAL/runterbeug_such_FRAU.baf": [
        "_DYNAMIC/Character/Holzfaellerin_FRAU.bgf"
    ],
    "Character/FRAU/SPEZIAL/gegenstand_schuetteln_zweihaendig_FRAU.baf": [
        "_DYNAMIC/Character/Handwerkerin3_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin3_FRAU.bgf",
        "_DYNAMIC/Character/Magd_FRAU.bgf",
        "_DYNAMIC/Character/Minerin2_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin_FRAU.bgf",
        "_DYNAMIC/Character/Patrizierin_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin2_FRAU.bgf",
        "_DYNAMIC/Character/Feldarbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Zigeunerin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin_FRAU.bgf",
        "_DYNAMIC/Character/Waschfrau_FRAU.bgf",
        "_DYNAMIC/Character/Minerin_FRAU.bgf",
        "_DYNAMIC/Character/Bauerin_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin3_FRAU.bgf",
        "_DYNAMIC/Character/Arbeiterin_FRAU.bgf",
        "_DYNAMIC/Character/Handwerkerin2_FRAU.bgf",
        "_DYNAMIC/Character/schneiderin1_FRAU.bgf",
        "_DYNAMIC/Character/Buergerin2_FRAU.bgf",
    ],
    "Character/MAEDCHEN/Bewegung/gehen_MAEDCHEN.baf": [],
    "Character/MAEDCHEN/Bewegung/dreh_45_links_MAEDCHEN.baf": [],
    "Character/MAEDCHEN/Bewegung/dreh_90_rechts_MAEDCHEN.baf": [],
    "Character/MAEDCHEN/Bewegung/dreh_90_links_MAEDCHEN.baf": [],
    "Character/MAEDCHEN/Bewegung/dreh_45_rechts_MAEDCHEN.baf": [],
    "Character/MAEDCHEN/Bewegung/step_MAEDCHEN.baf": [],
    "Character/Schwein/Bewegung/fressen_SCHWEIN.baf": [
        "_DYNAMIC/Tiere/schwein_SCHWEIN.bgf"
    ],
    "Character/Schwein/Bewegung/drehen_SCHWEIN.baf": [
        "_DYNAMIC/Tiere/schwein_SCHWEIN.bgf"
    ],
    "Character/Schwein/Bewegung/stehen_SCHWEIN.baf": [
        "_DYNAMIC/Tiere/schwein_SCHWEIN.bgf"
    ],
    "Character/Schwein/Bewegung/gehen_SCHWEIN.baf": [
        "_DYNAMIC/Tiere/schwein_SCHWEIN.bgf"
    ],
    "Character/MUTTER_BETT/mutter_bett.baf": ["_DYNAMIC/MUTTER_BETT/mutter_bett.bgf"],
    "Character/sensenmann/sensenmann.baf": ["_DYNAMIC/Character/Sensenmann.bgf"],
    "Character/JUNGE/Bewegung/dreh_45_rechts_JUNGE.baf": [],
    "Character/JUNGE/Bewegung/dreh_45_links_JUNGE.baf": [],
    "Character/JUNGE/Bewegung/dreh_90_links_JUNGE.baf": [],
    "Character/JUNGE/Bewegung/gehen_JUNGE.baf": [],
    "Character/JUNGE/Bewegung/step_JUNGE.baf": [],
    "Character/JUNGE/Bewegung/dreh_90_rechts_JUNGE.baf": [
        "_DYNAMIC/Gebaeude_Bauen/GASTHAUS_a/bau_GHA_pfeiler_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GASTHAUS_a/bau_GHA_pfeiler_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_d/bau_HD_pfahl_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_d/bau_HD_pfahl_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_b/bau_HB_pfahl_m_l_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_b/bau_HB_pfahl_m_r_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/DIEBESGILDE_a/bau_DIEB_PfaehleT_vr_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/DIEBESGILDE_a/bau_DIEB_Pfaehle_uhm_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/DIEBESGILDE_a/bau_DIEB_PfaehleT_vl_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/DIEBESGILDE_a/bau_DIEB_PfaehleT_hl_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_a/bau_HA_pfeiler_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_a/bau_HA_pfeiler_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_c/bau_HC_pfeiler_m_l_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_c/bau_HC_pfeiler_m_r_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Pfaehle_u_01.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Pfaehle_u_03.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Pfaehle_u_04.bgf",
        "_DYNAMIC/Gebaeude_Bauen/BRUNNEN_a/bau_BRUN_Pfaehle_u_02.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_lo_12.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_ho_10.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_hu_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_vu_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_ro_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_vo_09.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_ru_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_a/bauGefA_PfaehleK_lu_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_b/bauStWaB_Pfaehle_umh_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_b/bauStWaB_Pfaehle_uml_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_b/bauStWaB_Pfaehle_umv_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_b/bauStWaB_Pfaehle_umr_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_h/bau_HC_PfaehleK_vl_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_h/bau_HC_PfaehleK_hr_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_h/bau_HC_PfaehleK_hm_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_h/bau_HC_PfaehleK_hl_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_KLEIN/bau_KKA_dach_o_v_15.bgf",
        "_DYNAMIC/Gebaeude_Bauen/KIRCHE_KLEIN/bau_KKA_dach_o_m_16.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_f/bau_HF_pfeiler_mr_l_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_f/bau_HF_pfeiler_mr_r_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GELDLEIHE_a/bau_GELD_pfahl_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GELDLEIHE_a/bau_GELD_pfahl_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_a/bauStWaA_Pfaehle_urm_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_a/bauStWaA_Pfaehle_uvm_05.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_a/bauStWaA_Pfaehle_ulm_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/STADTWACHE_a/bauStWaA_Pfaehle_uhm_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_vsr_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_lm_06.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_vsl_10.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_vm_09.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_rm_08.bgf",
        "_DYNAMIC/Gebaeude_Bauen/HANDWERK_g/bau_HG_Pfaehle_hm_07.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_hu_03.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_vu_01.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_vo_09.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_lu_04.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_ro_10.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_ho_11.bgf",
        "_DYNAMIC/Gebaeude_Bauen/GEFAENGNIS_b/bauGefB_PfaehleK_ru_02.bgf",
    ],
    "BAUM/faellen_TANNE_02.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_01_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_04_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_03_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_02_FAELLT.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/HANDWERK_SCHNEIDEREI_b/bau_Hw_Schneider_02.bgf",
    ],
    "BAUM/wiegen_TANNE_KRONE.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
    ],
    "BAUM/faellen_BUCHE_04.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_03_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_02_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_04_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_01_FAELLT.bgf",
    ],
    "BAUM/faellen_BUCHE_03.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_03_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_02_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_04_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_01_FAELLT.bgf",
    ],
    "BAUM/faellen_EICHE_03.baf": [
        "Vegetation/EICHE/pfl_R_EICHE_03_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_01_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_04_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_02_FAELLT.bgf",
        "Gebaeude/gb_Schmiede/gb_SCHMIEDE_0.bgf",
        "Gebaeude/gb_Schmiede/gb_GOLDSCHMIEDE_0.bgf",
    ],
    "BAUM/zittern_BUCHE.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_03.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_02.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_01.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_04.bgf",
    ],
    "BAUM/hack_TANNE_KRONE.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
    ],
    "BAUM/flirren_BUCHE.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_03.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_02.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_01.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_KRONE_04.bgf",
    ],
    "BAUM/hack_BUCHE.baf": [],
    "BAUM/faellen_EICHE_02.baf": [
        "Vegetation/EICHE/pfl_R_EICHE_03_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_01_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_04_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_02_FAELLT.bgf",
        "Gebaeude/gb_Schmiede/gb_SCHMIEDE_0.bgf",
        "Gebaeude/gb_Schmiede/gb_GOLDSCHMIEDE_0.bgf",
    ],
    "BAUM/zittern_TANNE.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
    ],
    "BAUM/flirren_EICHE.baf": [
        "Vegetation/EICHE/pfl_R_EICHE_KRONE_01.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_KRONE_04.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_KRONE_02.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_KRONE_03.bgf",
    ],
    "BAUM/faellen_TANNE_01.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_01_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_04_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_03_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_02_FAELLT.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/HANDWERK_SCHNEIDEREI_b/bau_Hw_Schneider_02.bgf",
    ],
    "BAUM/faellen_EICHE_01.baf": [
        "Vegetation/EICHE/pfl_R_EICHE_03_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_01_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_04_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_02_FAELLT.bgf",
        "Gebaeude/gb_Schmiede/gb_SCHMIEDE_0.bgf",
        "Gebaeude/gb_Schmiede/gb_GOLDSCHMIEDE_0.bgf",
    ],
    "BAUM/flirren_LAUBKRONE.baf": [
        "Vegetation/Graeser/vg_GRAS_STAUDE2_3X1X1M_LAUBKRONE.bgf",
        "Vegetation/Buesche/vg_BUSCH_A_1X1X3M_LAUBKRONE.bgf",
        "Vegetation/Buesche/vg_BUSCH_A_3X2X1M_LAUBKRONE.bgf",
        "Vegetation/Buesche/vg_BUSCH_A_2X2X2M_LAUBKRONE.bgf",
    ],
    "BAUM/faellen_BUCHE_01.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_03_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_02_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_04_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_01_FAELLT.bgf",
    ],
    "BAUM/wiegen_TANNE_02.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
        "Accessoires/Felsen/Felsen_05_doppel.bgf",
        "Accessoires/Felsen/Felsen_10_doppel.bgf",
    ],
    "BAUM/wiegen_TANNE_04.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
        "Accessoires/Felsen/Felsen_05_doppel.bgf",
        "Accessoires/Felsen/Felsen_10_doppel.bgf",
    ],
    "BAUM/windhauch_LAUBKRONE01.baf": [
        "Vegetation/Laubbaeume/vg_LBAUM01_BREIT_X12KRONE_LAUBKRONE01.bgf",
        "Vegetation/Laubbaeume/vg_LBAUM01_HOCH_X7_KRONE_LAUBKRONE01.bgf",
        "Vegetation/Laubbaeume/vg_LBAUM01_HOCH_X12KRONE_LAUBKRONE01.bgf",
        "Vegetation/Laubbaeume/vg_LBAUM01_HOCH_X5_KRONE_LAUBKRONE01.bgf",
    ],
    "BAUM/faellen_TANNE_04.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_01_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_04_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_03_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_02_FAELLT.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/HANDWERK_SCHNEIDEREI_b/bau_Hw_Schneider_02.bgf",
    ],
    "BAUM/faellen_BUCHE_02.baf": [
        "Vegetation/BUCHE/pfl_R_BUCHE_03_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_02_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_04_FAELLT.bgf",
        "Vegetation/BUCHE/pfl_R_BUCHE_01_FAELLT.bgf",
    ],
    "BAUM/wiegen_TANNE_03.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
        "Accessoires/Felsen/Felsen_05_doppel.bgf",
        "Accessoires/Felsen/Felsen_10_doppel.bgf",
    ],
    "BAUM/wiegen_TANNE_01.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_03.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_02.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_04.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_KRONE_01.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_GRUEN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_MIX.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_10M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BRAUN.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_13M_BLAU.bgf",
        "Vegetation/Tannen/vg_TANNE_KRONE_06M_BRAUN.bgf",
        "Accessoires/Felsen/Felsen_05_doppel.bgf",
        "Accessoires/Felsen/Felsen_10_doppel.bgf",
    ],
    "BAUM/faellen_EICHE_04.baf": [
        "Vegetation/EICHE/pfl_R_EICHE_03_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_01_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_04_FAELLT.bgf",
        "Vegetation/EICHE/pfl_R_EICHE_02_FAELLT.bgf",
        "Gebaeude/gb_Schmiede/gb_SCHMIEDE_0.bgf",
        "Gebaeude/gb_Schmiede/gb_GOLDSCHMIEDE_0.bgf",
    ],
    "BAUM/faellen_TANNE_03.baf": [
        "Vegetation/TANNE/pfl_R_TANNE_01_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_04_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_03_FAELLT.bgf",
        "Vegetation/TANNE/pfl_R_TANNE_02_FAELLT.bgf",
        "_Addon/_DYNAMIC/Gebaeude_Bauen/HANDWERK_SCHNEIDEREI_b/bau_Hw_Schneider_02.bgf",
    ],
    "BAUM/zittern_EICHE.baf": [],
}
